from lib_cidr_trie import CIDRTrie
from lib_cidr_trie.cidr_trie import PrefixType

from .enums_and_dataclasses import ROAOutcome, ROARouted, ROAValidity
from .roa import ROA
from .roas_node import ROAsNode


class ROATrie(CIDRTrie[PrefixType]):
    """Trie of CIDRs for ROAs"""

    # Just changing the defaults here to ROA instead of CIDRNode
    def __init__(self, *args, NodeCls: type[ROAsNode] = ROAsNode, **kwargs) -> None:
        # mypy doesn't understand that I can reset the default arg in this way
        super(ROATrie, self).__init__(*args, NodeCls=NodeCls, **kwargs)  # type: ignore

    def get_relevant_roas(self, prefix: PrefixType) -> frozenset[ROA]:
        """Returns all relevant ROAs for a given prefix

        This is a bit non-intuitive, because you'd think that all ROAs relevant
        for a given prefix would be __at the prefix__. But this isn't the case.
        They could be higher up the trie, and simply have a __max length__ that
        is >= your prefix length. So you need to collect all ROAs all the way down,
        and for each ROA, determine if it's relevant, and then return them
        """

        self._validate_prefix(prefix)
        bits = self._get_binary_str_from_prefix(prefix)
        node = self.root
        roas: list[ROA] = list()
        for bit in bits[: prefix.prefixlen]:
            next_node = node.right if bool(int(bit)) else node.left
            if next_node is None:
                return frozenset(roas)
            elif next_node.prefix is not None:
                # Add ROAs if they are relevant
                assert isinstance(next_node, ROAsNode), "for mypy"
                for roa in next_node.roas:
                    if roa.covers_prefix(prefix):
                        roas.append(roa)
            node = next_node
        return frozenset(roas)

    def get_roa_outcome(self, prefix: PrefixType, origin: int) -> ROAOutcome:
        """Gets the validity and roa routed vs non rotued of a prefix-origin pair

        This can get fairly complicated, since there can be multiple ROAs
        for the same announcement, and each ROA can have a different validity.
        Essentially, we need to calculate the best validity for a given announcement
        and then that's the validity that should be used. I.e. the "most valid" roa
        is the one that should be used
        """

        relevant_roas = self.get_relevant_roas(prefix)

        if relevant_roas:
            # Return the best ROAOutcome
            rv = sorted([x.get_outcome(prefix, origin) for x in relevant_roas])[0]
            assert isinstance(rv, ROAOutcome), "for mypy"
            return rv
        else:
            return ROAOutcome(
                validity=ROAValidity.UNKNOWN, routed_status=ROARouted.UNKNOWN
            )



from dataclasses import dataclass
from enum import Enum


class ROAValidity(Enum):
    """ROAValidity values

    NOTE: it's possible that you could have two ROAs for
    the same prefix, each with different reasons why they are
    invalid. In that case, the ROAChecker returns the "best"
    validity (in the order below). It doesn't really matter,
    since they are both invalid anyways, and that's the only
    case where this conflict can occur
    """

    # NOTE: These values double as "scores" for validity,
    # so do NOT change the order
    # (used in the ROA class)
    VALID = 0
    UNKNOWN = 1
    INVALID_LENGTH = 2
    INVALID_ORIGIN = 3
    INVALID_LENGTH_AND_ORIGIN = 4

    @staticmethod
    def is_valid(roa_validity: "ROAValidity") -> bool:
        return roa_validity == ROAValidity.VALID

    @staticmethod
    def is_unknown(roa_validity: "ROAValidity") -> bool:
        return roa_validity == ROAValidity.UNKNOWN

    @staticmethod
    def is_invalid(roa_validity: "ROAValidity") -> bool:
        return roa_validity in (
            ROAValidity.INVALID_LENGTH,
            ROAValidity.INVALID_ORIGIN,
            ROAValidity.INVALID_LENGTH_AND_ORIGIN,
        )


class ROARouted(Enum):
    ROUTED = 0
    UNKNOWN = 1
    # A ROA is Non Routed if it is for an origin of ASN 0
    # This means that the prefix for this ROA should never be announced
    NON_ROUTED = 2


@dataclass(frozen=True, slots=True)
class ROAOutcome:
    validity: ROAValidity
    routed_status: ROARouted

    def __lt__(self, other) -> bool:
        if isinstance(other, ROAOutcome):
            return self.validity.value < other.validity.value
        else:
            return NotImplemented



from ipaddress import ip_network

from roa_checker import ROA, ROAChecker, ROAOutcome, ROARouted, ROAValidity


def test_tree():
    # TODO: Break up into unit tests
    trie = ROAChecker()
    cidrs = [ip_network(x) for x in ["1.2.0.0/16", "1.2.3.0/24", "1.2.3.4"]]
    routed_origin = 1
    for cidr in cidrs:
        trie.insert(cidr, ROA(cidr, routed_origin, cidr.prefixlen))
    for cidr in cidrs:
        outcome = trie.get_roa_outcome(cidr, routed_origin)
        assert outcome == ROAOutcome(ROAValidity.VALID, ROARouted.ROUTED)
        assert ROAValidity.is_unknown(outcome.validity) is False
        assert ROAValidity.is_invalid(outcome.validity) is False
        assert ROAValidity.is_valid(outcome.validity) is True

    non_routed_cidrs = [ip_network(x) for x in ["2.2.0.0/16", "2.2.3.0/24", "2.2.3.4"]]
    non_routed_origin = 0
    for cidr in non_routed_cidrs:
        trie.insert(cidr, ROA(cidr, non_routed_origin, cidr.prefixlen))
    for cidr in non_routed_cidrs:
        outcome = trie.get_roa_outcome(cidr, routed_origin)
        assert outcome == ROAOutcome(ROAValidity.INVALID_ORIGIN, ROARouted.NON_ROUTED)

    outcome = trie.get_roa_outcome(ip_network("1.0.0.0/8"), routed_origin)
    assert outcome.validity == ROAValidity.UNKNOWN
    assert outcome.routed_status == ROARouted.UNKNOWN
    outcome = trie.get_roa_outcome(ip_network("255.255.255.255"), routed_origin)
    assert outcome.validity == ROAValidity.UNKNOWN
    assert outcome.routed_status == ROARouted.UNKNOWN
    assert ROAValidity.is_unknown(outcome.validity) is True
    assert ROAValidity.is_invalid(outcome.validity) is False
    assert ROAValidity.is_valid(outcome.validity) is False
    outcome = trie.get_roa_outcome(ip_network("1.2.4.0/24"), routed_origin)
    assert outcome.validity == ROAValidity.INVALID_LENGTH
    assert outcome.routed_status == ROARouted.ROUTED
    assert ROAValidity.is_unknown(outcome.validity) is False
    assert ROAValidity.is_invalid(outcome.validity) is True
    assert ROAValidity.is_valid(outcome.validity) is False
    outcome = trie.get_roa_outcome(ip_network("1.2.3.0/24"), routed_origin + 1)
    assert outcome.validity == ROAValidity.INVALID_ORIGIN
    assert outcome.routed_status == ROARouted.ROUTED
    assert ROAValidity.is_unknown(outcome.validity) is False
    assert ROAValidity.is_invalid(outcome.validity) is True
    assert ROAValidity.is_valid(outcome.validity) is False
    outcome = trie.get_roa_outcome(ip_network("1.2.4.0/24"), routed_origin + 1)
    assert outcome.validity == ROAValidity.INVALID_LENGTH_AND_ORIGIN
    assert outcome.routed_status == ROARouted.ROUTED
    assert ROAValidity.is_unknown(outcome.validity) is False
    assert ROAValidity.is_invalid(outcome.validity) is True
    assert ROAValidity.is_valid(outcome.validity) is False
    outcome = trie.get_roa_outcome(ip_network("1.2.0.255"), routed_origin)
    assert outcome.validity == ROAValidity.INVALID_LENGTH
    assert outcome.routed_status == ROARouted.ROUTED
    outcome = trie.get_roa_outcome(ip_network("1.3.0.0/16"), routed_origin)
    assert outcome.validity == ROAValidity.UNKNOWN
    assert outcome.routed_status == ROARouted.UNKNOWN
    outcome = trie.get_roa_outcome(ip_network("1.2.0.255"), routed_origin)
    assert outcome.validity == ROAValidity.INVALID_LENGTH
    assert outcome.routed_status == ROARouted.ROUTED


def test_multiple_differing_roas():
    """Testing that all ROAs are considered

    This test has one less specific ROA that is valid with a long max length
    and one more specific roa that is invalid

    This should result in a valid ROA
    """

    # TODO: Break up into unit tests
    trie = ROAChecker()
    valid_ip_addr = ip_network("1.2.0.0/16")
    invalid_ip_addr = ip_network("1.2.3.0/24")
    trie.insert(valid_ip_addr, ROA(valid_ip_addr, 1, 24))
    trie.insert(invalid_ip_addr, ROA(invalid_ip_addr, 2, 24))
    assert trie.get_roa_outcome(invalid_ip_addr, 1).validity == ROAValidity.VALID


def test_get_roa_outcome_w_prefix_str_cached():
    trie = ROAChecker()
    valid_ip_addr = ip_network("1.2.0.0/16")
    invalid_ip_addr = ip_network("1.2.3.0/24")
    trie.insert(valid_ip_addr, ROA(valid_ip_addr, 1, 24))
    trie.insert(invalid_ip_addr, ROA(invalid_ip_addr, 2, 24))
    assert (
        trie.get_roa_outcome_w_prefix_str_cached(str(invalid_ip_addr), 1).validity
        == ROAValidity.VALID
    )
